/**
 * A state object that is passed to each retry policy function. A new RetryState
 * is initialized on each async operation call. An operation is considered to be
 * the entire sequence of attempts, including the first one.
 */
type RetryState = {
    /**
     * The number of attempts that have been made. Starts at 1, after invoking the
     * function for the first time.
     */
    attempt: number;
    /**
     * The delay applied to the last attempt. Starts at 0, before invoking the
     * function for the first time.
     */
    delay: number;
    /**
     * The time elapsed since the first attempt.
     */
    elapsed: number;
    /**
     * The error that was thrown on the last attempt.
     */
    error: unknown;
    /**
     * An array of all errors that have been thrown so far.
     */
    errors: unknown[];
    /**
     * The time when the first attempt was made.
     */
    start: number;
};
/**
 * A function to be called after each failed attempt to determine the delay
 * before the next attempt. The function should return a number of milliseconds
 * to wait before the next attempt, or throw an error to give up. You can also
 * use this function to implement custom retry logic, run side effects
 * (e.g. logging), add special handling for specific errors, wrap errors in
 * custom types, etc.
 *
 * The function will receive the current state of the retry operation, and the
 * next policy in the chain (if any). If provided, the policy can invoke the
 * next policy in the chain or not, if called, it can respect the delay value it
 * returned or return its own. In either case, when a policy function throws an
 * error, the async operation will be aborted and the error will be propagated
 * to the caller.
 */
type RetryPolicy = (state: Readonly<RetryState>, next?: RetryPolicy) => number;
type PolicyList = [RetryPolicy, RetryPolicy, ...(RetryPolicy | RetryPolicy[])[]] | [[RetryPolicy], RetryPolicy, ...(RetryPolicy | RetryPolicy[])[]] | [
    [
        RetryPolicy
    ],
    [
        RetryPolicy,
        ...RetryPolicy[]
    ],
    ...(RetryPolicy | RetryPolicy[])[]
] | [
    RetryPolicy,
    [
        RetryPolicy,
        ...RetryPolicy[]
    ],
    ...(RetryPolicy | RetryPolicy[])[]
] | [
    [
        RetryPolicy,
        RetryPolicy,
        ...RetryPolicy[]
    ],
    ...(RetryPolicy | RetryPolicy[])[]
];
/**
 * Joins multiple retry policies into a single policy. The policies will be
 * called from left to right, each with the next policy as the second argument.
 *
 * @param policies The policies to join.
 * @returns A new policy that is the result of joining the provided policies.
 */
declare const join: (...policies: PolicyList) => RetryPolicy;
type AbortWrapper<Args extends unknown[], Input extends unknown[], Output> = Args extends [AbortSignal] ? (...args: Input) => Promise<Output> : Promise<Output>;
type WrappedFunction<Input extends unknown[], Output> = <Args extends [AbortSignal] | Input>(...args: Args) => AbortWrapper<Args, Input, Output>;
/**
 * Wrap an async function with a retry policy. The returned function has the
 * same signature as the input function, but will retry the operation according
 * to the provided policy if an error is thrown.
 *
 * The returned function is overloaded to accept an AbortSignal as its only
 * argument, in such case, a function is returned with the original signature.
 *
 * @example
 * import { core as wrap } from 'retryyy/core'
 * import type RetryPolicy from 'retryyy/core'
 *
 * const simpleExamplePolicy: RetryPolicy = ({ attempt, error }) => {
 * 	// Give up after 3 tries.
 * 	if (attempt > 3) {
 * 		throw error
 * 	}
 *
 * 	// Linear backoff, waits 1s, 2s, 3s, 4s, etc.
 * 	return attempt * 1000
 * }
 *
 * type UserShape = { id: number; name: string }
 *
 * export const fetchUser = wrap(async (id: number) => {
 * 	const res = await fetch(`https://jsonplaceholder.typicode.com/users/${id}`)
 * 	return await res.json() as UserShape
 * }, simpleExamplePolicy)
 */
declare const core: <Input extends unknown[], Output>(fn: (...input: Input) => Promise<Output>, policy: RetryPolicy) => WrappedFunction<Input, Output>;

export { type AbortWrapper, type PolicyList, type RetryPolicy, type RetryState, type WrappedFunction, core, join };
