import { RetryPolicy } from '../core.js';

type DefaultOptions = {
    /**
     * If true, runs the first re-attempt immediately, skipping the initial delay.
     * @default false
     */
    fastTrack?: boolean | undefined;
    /**
     * The initial delay in milliseconds. Defaults to 150ms.
     * @default 150
     */
    initialDelay?: number | undefined;
    /**
     * Logger function to use when giving up on retries. Passing `false` disables this type of
     * logging. Defaults to `console.error`.
     * @default console.error
     */
    logError?: typeof console.error | boolean | undefined;
    /**
     * Logger function to use when retrying. Passing `false` disables this type of logging. Defaults
     * to `console.warn`.
     * @default console.warn
     */
    logWarn?: typeof console.warn | boolean | undefined;
    /**
     * The maximum number of attempts to make. Defaults to 10.
     * @default 10
     */
    maxAttempts?: number | undefined;
    /**
     * The maximum delay between attempts in milliseconds. Defaults to 30 seconds.
     * @default 30000
     */
    maxDelay?: number | undefined;
    /**
     * The time in milliseconds after which to give up. Defaults to 30 seconds.
     * @default 30000
     */
    timeout?: number | undefined;
    /**
     * Convenience property to chain another policy after the default ones. This
     * is useful when you want to add custom logic on top of the default policies.
     * This is equivalent to `policy = join(Default(), next)`.
     */
    next?: RetryPolicy;
};
/**
 * The default retry policy used by `retryyy`. It includes:
 *
 * - Logger: logs retry attempts and failed operations to the console.
 * - Timeout: give up after a certain amount of time has passed.
 * - Breaker: stop retrying after a certain number of attempts.
 * - FastTrack: run the first re-attempt immediately if `fastTrack` is `true`.
 * - PollyJitter: advanced exponential backoff strategy with jitter.
 */
declare const Default: (options?: DefaultOptions) => RetryPolicy;

export { Default, type DefaultOptions };
