import { RetryPolicy, WrappedFunction } from './core.js';
export { AbortWrapper, PolicyList, RetryState, core, join } from './core.js';
import { DefaultOptions } from './policies/Default.js';
export { Default } from './policies/Default.js';
export { Backoff, BackoffOptions } from './policies/Backoff.js';
export { BrandError, RetryyyError } from './policies/BrandError.js';
export { Breaker, BreakerOptions } from './policies/Breaker.js';
export { FastTrack } from './policies/FastTrack.js';
export { DecorrelatedJitter, DecorrelatedJitterOptions, EqualJitter, FullJitter, Jitter, JitterOptions, PollyJitter, PollyJitterOptions } from './policies/Jitter.js';
export { Logger, LoggerOptions } from './policies/Logger.js';
export { Timeout, TimeoutOptions } from './policies/Timeout.js';

/**
 * Wrap an async function with a retry policy. The returned function has the
 * same signature as the input function, but will retry the operation according
 * to the provided policy if an error is thrown.
 *
 * The returned function is overloaded to accept an AbortSignal as its only
 * argument, in such case, a function is returned with the original signature.
 *
 * @example
 * import { wrap } from 'retryyy'
 *
 * type UserShape = { id: number; name: string }
 *
 * async function _fetchUser(id: number): Promise<UserShape> {
 * 	const res = await fetch(`https://jsonplaceholder.typicode.com/users/${id}`)
 * 	return await res.json() as UserShape
 * }
 *
 * export const fetchUser = wrap(_fetchUser, { timeout: 10_000 })
 */
declare const wrap: <Input extends unknown[], Output>(fn: (...args: Input) => Promise<Output>, policy?: RetryPolicy | DefaultOptions) => WrappedFunction<Input, Output>;
/**
 * Retries an async function based on a retry policy.
 *
 * @param fn Async function to run and retry, if needed.
 * @param policy Function that drives the retry logic.
 * @param signal Optional signal to abort the retry loop.
 */
declare const retryyy: <Output>(fn: () => Promise<Output>, policy?: RetryPolicy | DefaultOptions, signal?: AbortSignal) => Promise<Output>;
/**
 * Decorator for class methods that retries the method based on a retry policy.
 */
declare const Retryyy: (policy?: RetryPolicy | DefaultOptions) => <This, Input extends unknown[], Output>(target: (this: This, ...args: Input) => Promise<Output>) => WrappedFunction<Input, Output>;

export { DefaultOptions, RetryPolicy, Retryyy, WrappedFunction, retryyy, wrap };
